#open "support";;
#open "protocol";;
#open "textvariable";;
(* Report globals from protocol to tk *)
let OpenTk = OpenTk
and OpenTkClass = OpenTkClass
and CloseTk = CloseTk
and MainLoop = MainLoop
;;

let add_fileinput = add_fileinput
and remove_fileinput = remove_fileinput
;;
(* Tk_GetBitmap emulation *)
type Bitmap =
   BitmapFile of string
 | Predefined of string
;;

let CAMLtoTKBitmap = function
  BitmapFile s -> "\"@" ^ s ^ "\""
| Predefined s -> quote_string s
;;

let TKtoCAMLBitmap s = 
 if nth_char s 0 = `@`
 then BitmapFile (sub_string s 1 (string_length s - 1))
 else Predefined s
;;


(* Color *)
type Color =
	NamedColor of string
	| Black
	| White
	| Red
	| Green
	| Blue
;;

let CAMLtoTKColor = function
	NamedColor x -> quote_string x
	| Black -> "black"
	| White -> "white"
	| Red -> "red"
	| Green -> "green"
	| Blue -> "blue"
;;

let TKtoCAMLColor = function  s -> NamedColor s
;;


(* Tk_GetCursor emulation *)
type Cursor =
   XCursor of string 
 | XCursorFg of string * Color
 | XCursortFgBg of string * Color * Color
 | CursorFileFg of string * Color 
 | CursorMaskFile of string * string * Color * Color
;;

let CAMLtoTKCursor = function
   XCursor s -> quote_string s
 | XCursorFg (s,fg) -> 
    "[list " ^quote_string s^" "^(CAMLtoTKColor fg)^"]"
 | XCursortFgBg (s,fg,bg) ->
    "[list "^quote_string s^" "^(CAMLtoTKColor fg)^" "^(CAMLtoTKColor bg)^"]"
 | CursorFileFg (s,fg) ->
    "[list @\""^s^"\" "^(CAMLtoTKColor fg)^"]"
 | CursorMaskFile (s,m,fg,bg) ->
    "[list @\""^s^"\" \""^m^"\" "^(CAMLtoTKColor fg)^(CAMLtoTKColor bg)^"]"
;;


(* Tk_GetPixels emulation *)
type Units =
    Pixels of int       (* specified as floating-point, but inconvenient *)
  | Centimeters of float
  | Inches of float
  | Millimeters of float
  | PrinterPoint of float
;;

let CAMLtoTKUnits = function
    Pixels (foo) -> (string_of_int foo)
  | Millimeters (foo)  -> (string_of_float foo)^"m"
  | Inches (foo)  -> (string_of_float foo)^"i"
  | PrinterPoint (foo) -> (string_of_float foo)^"p"
  | Centimeters (foo) -> (string_of_float foo)^"c"
  ;;

let TKtoCAMLUnits str = 
  let len = string_length str in
  let num_part str = sub_string str 0 (len - 1) in
  match nth_char str (pred len) with
    `c` -> Centimeters (float_of_string (num_part str))
  | `i` -> Inches (float_of_string (num_part str))
  | `m` -> Millimeters (float_of_string (num_part str))
  | `p` -> PrinterPoint (float_of_string (num_part str))
  | _ -> Pixels(int_of_string str)
;;
(* Events and bindings *)

(* Builtin types *)
type XEvent =
    XKey of string	(* /usr/include/X11/keysymdef.h *)
  | ButtonPress
  | Button
  | ButtonRelease
  | Circulate
  | CirculateRequest
  | ColorMap
  | Configure
  | ConfigureRequest
  | Destroy
  | Enter
  | Expose
  | FocusIn
  | FocusOut
  | Gravity
  | Keymap
  | KeyPress
  | Key
  | KeyRelease
  | MapRequest
  | Motion
  | Leave
  | Map
  | Property
  | Reparent
  | ResizeRequest
  | Unmap
  | Visibility 
  | WhatButton of int
;;

let CAMLtoTKXEvent = function
    XKey a -> a
  | ButtonPress -> "ButtonPress"
  | Button -> "Button"
  | ButtonRelease -> "ButtonRelease"
  | Circulate -> "Circulate"
  | CirculateRequest -> "CirculateRequest"
  | ColorMap -> "ColorMap"
  | Configure -> "Configure"
  | ConfigureRequest -> "ConfigureRequest"
  | Destroy -> "Destroy"
  | Enter -> "Enter"
  | Expose -> "Expose"
  | FocusIn -> "FocusIn"
  | FocusOut -> "FocusOut"
  | Gravity -> "Gravity"
  | Keymap -> "Keymap"
  | KeyPress -> "KeyPress"
  | Key -> "Key"
  | KeyRelease -> "KeyRelease"
  | MapRequest -> "MapRequest"
  | Motion -> "Motion"
  | Leave -> "Leave"
  | Map -> "Map"
  | Property -> "Property"
  | Reparent -> "Reparent"
  | ResizeRequest -> "ResizeRequest"
  | Unmap -> "Unmap"
  | Visibility -> "Visibility" 
  | WhatButton n -> string_of_int n

;;

type Modifier =
    Control
  | Shift
  | Lock
  | Button1
  | Button2
  | Button3
  | Button4
  | Button5
  | Any
  | Double
  | Triple
  | Mod1
  | Mod2
  | Mod3
  | Mod4
  | Mod5
  | Meta
  | Alt ;;


let CAMLtoTKModifier = function
   Control ->"Control-"
 | Shift -> "Shift-"
 | Lock -> "Lock-"
 | Button1 -> "Button1-"
 | Button2 -> "Button2-"
 | Button3 -> "Button3-"
 | Button4 -> "Button4-"
 | Button5 -> "Button5-"
 | Any -> "Any-"
 | Double -> "Double-"
 | Triple -> "Triple-"
 | Mod1 -> "Mod1-"
 | Mod2 -> "Mod2-"
 | Mod3 -> "Mod3-"
 | Mod4 -> "Mod4-"
 | Mod5 -> "Mod5-"
 | Meta -> "Meta-"
 | Alt -> "Alt-"
;;


(* type Event = Modifier list * XEvent *)
let CAMLtoTKEvent (ml, xe) =
  "<" ^ (catenate_sep " " (map CAMLtoTKModifier ml))  
      ^ (CAMLtoTKXEvent xe) ^ ">"
;;
  
(* type EventSequence == (Modifier list * XEvent) list *)
let CAMLtoTKEventSequence l =
  it_list (prefix ^) "" (map CAMLtoTKEvent l)
;;

(* Event structure, passed to bounded functions *)

type EventInfo = {
  mutable Ev_Above : int;
  mutable Ev_ButtonNumber : int;
  mutable Ev_Count : int;
  mutable Ev_Detail : string;
  mutable Ev_Focus : bool;
  mutable Ev_Height : int;
  mutable Ev_KeyCode : int;
  mutable Ev_Mode : string;
  mutable Ev_State : string;
  mutable Ev_ValueMask : int;
  mutable Ev_Width : int;
  mutable Ev_MouseX : int;
  mutable Ev_MouseY : int;
  mutable Ev_WidgetName : string;
  mutable Ev_Time : string;
  mutable Ev_OverrideRedirect : string;
  mutable Ev_Place : string;
  mutable Ev_Char : string;
  mutable Ev_BorderWidth : int;
  mutable Ev_Display : int;
  mutable Ev_SendEvent : int;
  mutable Ev_KeySymString : string;
  mutable Ev_KeySymInt : int;
  mutable Ev_RootWindow : int;
  mutable Ev_SubWindow : int;
  mutable Ev_Type : int;
  mutable Ev_RootX : int;
  mutable Ev_RootY : int };;


type EventField =
    Above
  | ButtonNumber
  | Count
  | Detail
  | Focus
  | Height
  | KeyCode
  | Mode
  | State
  | ValueMask
  | Width
  | MouseX
  | MouseY
  | WidgetName
  | Time 
  | OverrideRedirect
  | Place
  | Char
  | BorderWidth
  | Display
  | SendEvent
  | KeySymString
  | KeySymInt
  | RootWindow
  | SubWindow
  | Type
  | RootX
  | RootY;;

(* TODO: check types ! *)

let FillEventInfo ev args = function 
    Above 	-> ev.Ev_Above 		<- int_of_string (arg_GetTkToken args)
  | ButtonNumber -> ev.Ev_ButtonNumber 	<- int_of_string (arg_GetTkToken args)
  | Count 	-> ev.Ev_Count 		<- int_of_string (arg_GetTkToken args)
  | Detail 	-> ev.Ev_Detail 	<- arg_GetTkToken args
  | Focus 	-> ev.Ev_Focus 		<- arg_GetTkToken args = "1"
  | Height 	-> ev.Ev_Height 	<- int_of_string (arg_GetTkToken args)
  | KeyCode 	-> ev.Ev_KeyCode 	<- int_of_string (arg_GetTkToken args)
  | Mode 	-> ev.Ev_Mode 		<- arg_GetTkToken args
  | State 	-> ev.Ev_State 		<- arg_GetTkToken args
  | ValueMask	-> ev.Ev_ValueMask 	<- int_of_string (arg_GetTkToken args)
  | Width 	-> ev.Ev_Width 		<- int_of_string (arg_GetTkToken args)
  | MouseX 	-> ev.Ev_MouseX 	<- int_of_string (arg_GetTkToken args)
  | MouseY 	-> ev.Ev_MouseY 	<- int_of_string (arg_GetTkToken args)
  | WidgetName 	-> ev.Ev_WidgetName 	<- arg_GetTkToken args
  | Time 	-> ev.Ev_Time  		<- arg_GetTkToken args
  | OverrideRedirect -> ev.Ev_OverrideRedirect <- arg_GetTkToken args
  | Place 	-> ev.Ev_Place 		<- arg_GetTkToken args
  | Char 	-> ev.Ev_Char 		<- arg_GetTkToken args (* WARNING *)
  | BorderWidth -> ev.Ev_BorderWidth 	<- int_of_string (arg_GetTkToken args)
  | Display 	-> ev.Ev_Display 	<- int_of_string (arg_GetTkToken args)
  | SendEvent 	-> ev.Ev_SendEvent 	<- int_of_string (arg_GetTkToken args)
  | KeySymString -> ev.Ev_KeySymString 	<- arg_GetTkToken args
  | KeySymInt 	-> ev.Ev_KeySymInt 	<- int_of_string (arg_GetTkToken args)
  | RootWindow 	-> ev.Ev_RootWindow 	<- int_of_string (arg_GetTkToken args)
  | SubWindow 	-> ev.Ev_SubWindow 	<- int_of_string (arg_GetTkToken args)
  | Type 	-> ev.Ev_Type 		<- int_of_string (arg_GetTkToken args)
  | RootX 	-> ev.Ev_RootX 		<- int_of_string (arg_GetTkToken args)
  | RootY 	-> ev.Ev_RootY 		<- int_of_string (arg_GetTkToken args)
;;


let WrapEventInfo f what =
  let ev = {
    Ev_Above = 0;
    Ev_ButtonNumber = 0;
    Ev_Count = 0;
    Ev_Detail = "";
    Ev_Focus = false;
    Ev_Height = 0;
    Ev_KeyCode = 0;
    Ev_Mode = "";
    Ev_State = "";
    Ev_ValueMask = 0;
    Ev_Width = 0;
    Ev_MouseX = 0;
    Ev_MouseY = 0;
    Ev_WidgetName = "";
    Ev_Time = "";
    Ev_OverrideRedirect = "";
    Ev_Place = "";
    Ev_Char = "";
    Ev_BorderWidth = 0;
    Ev_Display = 0;
    Ev_SendEvent = 0;
    Ev_KeySymString = "";
    Ev_KeySymInt = 0;
    Ev_RootWindow = 0;
    Ev_SubWindow = 0;
    Ev_Type = 0;
    Ev_RootX = 0;
    Ev_RootY = 0 } in
     function args ->
       do_list (FillEventInfo ev args) what;
       f ev
;;



let rec WriteEventField = function
    [] -> ""
  | H::q ->
    begin
    match H with
        Above -> 	" %a"
      | ButtonNumber -> " %b"
      | Count -> 	" %c"
      | Detail -> 	" %d"
      | Focus -> 	" %f"
      | Height -> 	" %h"
      | KeyCode -> 	" %k"
      | Mode -> 	" %m"
      | State -> 	" %s"
      | ValueMask -> 	" %v"
      | Width -> 	" %w"
      | MouseX -> 	" %x"
      | MouseY -> 	" %y"
      | WidgetName -> 	" %W"
      | Time -> 	" %t"
      | OverrideRedirect -> " %o"
      | Place -> 	" %p"
      (* Quoting is done by Tk *)
      | Char -> 	" %A"
      | BorderWidth -> 	" %B"
      | Display -> 	" %D"
      | SendEvent -> 	" %E"
      | KeySymString -> " %K"
      | KeySymInt -> 	" %N"
      | RootWindow -> 	" %R"
      | SubWindow -> 	" %S"
      | Type -> 	" %T"
      | RootX -> 	" %X"
      | RootY -> 	" %Y"
    end 
    ^ WriteEventField q;;


type BindAction =
   BindSet of EventField list *  (EventInfo -> unit)
 | BindRemove
 | BindExtend of EventField list *  (EventInfo -> unit)
;;


(* bind: Widget -> (Modifier list * XEvent) list -> BindAction -> unit *)

let bind widget eventsequence action =
  let buf = Send2TkStart false in
  Send2Tk buf ("bind "^ widget_name widget^" "
      	              ^ (CAMLtoTKEventSequence eventsequence));
  begin match action with
     BindRemove -> Send2Tk buf "{}"
  |  BindSet (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  |  BindExtend (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {+camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  end;
  Send2TkEval buf;
  ()
;;

(* class_bind : string -> (Modifier list * XEvent) list -> BindAction -> unit 
      class arg is not constrained *)
let class_bind class eventsequence action =
  let buf = Send2TkStart false in
  Send2Tk buf ("bind "^ class ^" "
      	              ^ (CAMLtoTKEventSequence eventsequence));
  begin match action with
     BindRemove -> Send2Tk buf "{}"
  |  BindSet (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  |  BindExtend (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {+camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  end;
  Send2TkEval buf;
  ()
;;
(* TagOrId is builtin  because of SearchSpec *)
type TagOrId =
    Tag of string
  | Id of int ;;


let CAMLtoTKTagOrId = function
   Tag name -> name	
 | Id  n -> string_of_int n
;;

let TKtoCAMLTagOrId n = 
  try 
    Id (int_of_string n) 
  with
    _ -> Tag n
;;

type SearchSpec =
    Above of TagOrId
  | All
  | Below of TagOrId
  | Closest of int * int
  | Enclosed of int * int * int * int
  | Overlapping of int * int * int * int
  | Withtag of TagOrId
;;

let CAMLtoTKSearchSpec = function
    Above a -> "above "^(CAMLtoTKTagOrId a)
  | All -> "all "
  | Below a -> "below "^(CAMLtoTKTagOrId a)
  | Closest (a,b) -> "closest "^(string_of_int a)^" "^(string_of_int b)
  | Enclosed (a,b,c,d) -> 
      	"enclosed "^(string_of_int a)^" "^(string_of_int b)^" "
        ^(string_of_int c)^" "^(string_of_int d)
  | Overlapping (a,b,c,d) -> 
      	"overlapping "^(string_of_int a)^" "^(string_of_int b)^" "
        ^(string_of_int c)^" "^(string_of_int d)
  | Withtag a -> "withtag "^(CAMLtoTKTagOrId a)
;;


type CanvasIndex = 
    CI_Number of int
  | CI_End
  | CI_Insert
  | CI_SelFisrt
  | CI_SelLast
  | CI_At of int * int
;;


let CAMLtoTKCanvasIndex = function
    CI_Number (bar) -> string_of_int bar
  | CI_End -> "end"
  | CI_Insert -> "insert"
  | CI_SelFisrt -> "sel.first"
  | CI_SelLast -> "sel.last"
  | CI_At (foo, bar) -> ("@"^(string_of_int foo)^","^(string_of_int bar))
;;


(* TODO: restrict event fields *)
let canvas_bind widget tag eventsequence action =
  check_widget_class widget "canvas";
  let buf = Send2TkStart false in
  Send2Tk buf (widget_name widget ^ " bind " ^ (CAMLtoTKTagOrId tag) ^ " " ^
      	   (CAMLtoTKEventSequence eventsequence));
  begin match action with
     BindRemove -> Send2Tk buf "{}"
  |  BindSet (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  |  BindExtend (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {+camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  end;
  Send2TkEval buf
;;

type MenuIndex =
    MI_Number of int
  | MI_Active
  | MI_Last
  | MI_None
  | MI_At of int
  | MI_Pattern of string
;;


let CAMLtoTKMenuIndex = function
    MI_Active -> "active"
  | MI_Last -> "last"
  | MI_None -> "none"
  | MI_At (foo) -> "@"^(string_of_int foo)
  | MI_Number (foo) -> string_of_int foo
  | MI_Pattern (foo) -> foo
;;
(* Not a string as such, more like a symbol *)

type TextMark == string ;;

let CAMLtoTKTextMark  x =  x
;;
let TKtoCAMLTextMark x = x
;;

type TextTag == string ;;

let CAMLtoTKTextTag  x =  x
;;
let TKtoCAMLTextTag x = x
;;


type TextModifier =
    CharOffset of int
  | LineOffset of int
  | LineStart
  | LineEnd
  | WordStart
  | WordEnd 
;;

(* TextModifiers are never returned by Tk *)
let ppTextModifier = function
   CharOffset n -> 
      if n > 0 then "+" ^ (string_of_int n) ^ "chars"
      else if n = 0 then ""
      else (string_of_int n) ^ "chars"
 | LineOffset n -> 
      if n > 0 then "+" ^ (string_of_int n) ^ "lines"
      else if n = 0 then ""
      else (string_of_int n) ^ "lines"
 | LineStart -> " linestart"
 | LineEnd -> " lineend"
 | WordStart -> " wordstart"
 | WordEnd -> " wordend"
;;

type BaseTextIndex =
   TI_LineChar of int * int
 | TI_Mark of TextMark
 | TI_End 
 | TI_At of int * int
 | TI_TagFirst of TextTag
 | TI_TagLast of TextTag
;;

let CAMLtoTKBaseTextIndex = function
   TI_LineChar (l,c) -> (string_of_int l) ^ "." ^ (string_of_int c)
 | TI_Mark s -> s
 | TI_End -> "end"
 | TI_At(x,y) -> "@" ^ (string_of_int x) ^ "," ^ (string_of_int y)
 | TI_TagFirst s -> s ^ ".first"
 | TI_TagLast s -> s ^ ".last"
;;


let char_index c s = find 0
  where rec find i =
    if i >= string_length s 
    then raise Not_found
    else if nth_char s i = c then i 
    else find (i+1) 
;;


(* Assumes only l.c is returned *)
let TKtoCAMLBaseTextIndex s = 
  try
   let p = char_index `.` s in
     TI_LineChar(int_of_string (sub_string s 0 p), 
      	         int_of_string (sub_string s (p+1) (string_length s - p - 1)))
  with 
    Not_found -> raise (Invalid_argument "TKtoCAMLBaseTextIndex")
;;


type TextIndex =
   TextIndex of BaseTextIndex * TextModifier list
 | TextIndexNone
;;

let CAMLtoTKTextIndex = function
   TextIndexNone -> ""
 | TextIndex (base, ml) -> 
     "\"" ^ 
     it_list (prefix ^) (CAMLtoTKBaseTextIndex base) (map ppTextModifier ml)
     ^ "\""
;;


let text_tag_bind widget tag eventsequence action =
  check_widget_class widget "text";
  let buf = Send2TkStart false in
  Send2Tk buf (widget_name widget ^ " tag bind " ^ (CAMLtoTKTextTag tag) ^ " " ^
      	   (CAMLtoTKEventSequence eventsequence));
  begin match action with
     BindRemove -> Send2Tk buf "{}"
  |  BindSet (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  |  BindExtend (what, f) ->
      let CbId = register_callback (WrapEventInfo f what) in
        Send2Tk buf (" {+camlcb " ^ CbId ^ (WriteEventField what) ^"}")
  end;
  Send2TkEval buf;
  ()
;;
type Relief =
	Groove
	| Ridge
	| Flat
	| Sunken
	| Raised
;;

let CAMLtoTKRelief = function
	Groove -> "groove"
	| Ridge -> "ridge"
	| Flat -> "flat"
	| Sunken -> "sunken"
	| Raised -> "raised"
;;

type Orientation =
	Horizontal
	| Vertical
;;

let CAMLtoTKOrientation = function
	Horizontal -> "horizontal"
	| Vertical -> "vertical"
;;

type Anchor =
	SE
	| S
	| SW
	| E
	| Center
	| W
	| NE
	| N
	| NW
;;

let CAMLtoTKAnchor = function
	SE -> "se"
	| S -> "s"
	| SW -> "sw"
	| E -> "e"
	| Center -> "center"
	| W -> "w"
	| NE -> "ne"
	| N -> "n"
	| NW -> "nw"
;;

type State =
	Disabled
	| Active
	| Normal
;;

let CAMLtoTKState = function
	Disabled -> "disabled"
	| Active -> "active"
	| Normal -> "normal"
;;

type ArcStyle =
	Arc
	| Chord
	| Pieslice
;;

let CAMLtoTKArcStyle = function
	Arc -> "arc"
	| Chord -> "chord"
	| Pieslice -> "pieslice"
;;

type JoinStyle =
	JS_Round
	| JS_Miter
	| JS_Bevel
;;

let CAMLtoTKJoinStyle = function
	JS_Round -> "round"
	| JS_Miter -> "miter"
	| JS_Bevel -> "bevel"
;;

type CapStyle =
	CS_Round
	| CS_Projecting
	| CS_Butt
;;

let CAMLtoTKCapStyle = function
	CS_Round -> "round"
	| CS_Projecting -> "projecting"
	| CS_Butt -> "butt"
;;

type ArrowStyle =
	Arrow_Both
	| Arrow_Last
	| Arrow_First
	| Arrow_None
;;

let CAMLtoTKArrowStyle = function
	Arrow_Both -> "both"
	| Arrow_Last -> "last"
	| Arrow_First -> "first"
	| Arrow_None -> "none"
;;

type Justification =
	Justify_Right
	| Justify_Center
	| Justify_Left
;;

let CAMLtoTKJustification = function
	Justify_Right -> "right"
	| Justify_Center -> "center"
	| Justify_Left -> "left"
;;

type ColorMode =
	Mono
	| Gray
	| Color
;;

let CAMLtoTKColorMode = function
	Mono -> "mono"
	| Gray -> "gray"
	| Color -> "color"
;;

type WrapMode =
	WrapWord
	| WrapChar
	| WrapNone
;;

let CAMLtoTKWrapMode = function
	WrapWord -> "word"
	| WrapChar -> "char"
	| WrapNone -> "none"
;;

type Side =
	Side_Bottom
	| Side_Top
	| Side_Right
	| Side_Left
;;

let CAMLtoTKSide = function
	Side_Bottom -> "bottom"
	| Side_Top -> "top"
	| Side_Right -> "right"
	| Side_Left -> "left"
;;

type FillMode =
	Fill_Both
	| Fill_Y
	| Fill_X
	| Fill_None
;;

let CAMLtoTKFillMode = function
	Fill_Both -> "both"
	| Fill_Y -> "y"
	| Fill_X -> "x"
	| Fill_None -> "none"
;;

type BorderMode =
	BorderIgnore
	| BorderOutside
	| BorderInside
;;

let CAMLtoTKBorderMode = function
	BorderIgnore -> "ignore"
	| BorderOutside -> "outside"
	| BorderInside -> "inside"
;;

type option =
	RelX of Units
	| RelY of Units
	| RelWidth of float
	| RelHeigth of float
	| BorderMode of BorderMode
	| After of Widget
	| Before of Widget
	| Expand of bool
	| Fill of FillMode
	| In of Widget
	| IPadX of Units
	| IPadY of Units
	| Side of Side
	| Screen of string
	| Wrap of WrapMode
	| TextHeight of int
	| BgStipple of Bitmap
	| FgStipple of Bitmap
	| Underline of bool
	| To of int
	| TickInterval of int
	| SliderLength of Units
	| SliderForeground of Color
	| ShowValue of bool
	| Length of Units
	| From of int
	| Slidecommand of int -> unit
	| Aspect of int
	| Accelerator of string
	| Label of string
	| PostCommand of unit -> unit
	| Menu of Widget
	| ScrollRegion of Units * Units * Units * Units
	| ScrollIncrement of Units
	| Confine of bool
	| CloseEnough of float
	| Colormap of string
	| Colormode of ColorMode
	| Fontmap of string
	| PageAnchor of Anchor
	| PageHeight of Units
	| PageWidth of Units
	| PageX of Units
	| PageY of Units
	| Rotate of bool
	| X of Units
	| Y of Units
	| Window of Widget
	| Justify of Justification
	| ArrowStyle of ArrowStyle
	| ArrowShape of Units * Units * Units
	| CapStyle of CapStyle
	| JoinStyle of JoinStyle
	| Smooth of bool
	| SplineSteps of int
	| Extent of float
	| FillColor of Color
	| Outline of Color
	| Start of float
	| Stipple of Bitmap
	| ArcStyle of ArcStyle
	| Tags of TagOrId list
	| TextWidth of int
	| Value of string
	| Variable of TextVariable
	| Selector of Color
	| OnValue of string
	| OffValue of string
	| State of State
	| Command of unit -> unit
	| Class of string
	| Width of Units
	| Height of Units
	| ActiveBackground of Color
	| ActiveBorderWidth of Units
	| ActiveForeground of Color
	| Anchor of Anchor
	| Background of Color
	| Bitmap of Bitmap
	| Borderwidth of Units
	| Cursor of Cursor
	| CursorBackground of Color
	| CursorBorderWidth of Units
	| CursorOffTime of int
	| CursorOnTime of int
	| CursorWidth of Units
	| DisabledForeground of Color
	| ExportSelection of bool
	| Font of string
	| Foreground of Color
	| Geometry of string
	| InsertBackground of Color
	| InsertBorderWidth of Units
	| InsertOffTime of int
	| InsertOnTime of int
	| InsertWidth of Units
	| Orientation of Orientation
	| PadX of Units
	| PadY of Units
	| Relief of Relief
	| RepeatDelay of int
	| RepeatInterval of int
	| ScrollCommand of int -> int -> int -> int ->  unit
	| SelectBackground of Color
	| SelectBorderWidth of Units
	| SelectForeground of Color
	| SetGrid of bool
	| Text of string
	| TextVariable of TextVariable
	| UnderlinedChar of int
	| XScrollCommand of int -> int -> int -> int ->  unit
	| YScrollCommand of int -> int -> int -> int ->  unit
;;

type option_constrs =
	CRelX
	| CRelY
	| CRelWidth
	| CRelHeigth
	| CBorderMode
	| CAfter
	| CBefore
	| CExpand
	| CFill
	| CIn
	| CIPadX
	| CIPadY
	| CSide
	| CScreen
	| CWrap
	| CTextHeight
	| CBgStipple
	| CFgStipple
	| CUnderline
	| CTo
	| CTickInterval
	| CSliderLength
	| CSliderForeground
	| CShowValue
	| CLength
	| CFrom
	| CSlidecommand
	| CAspect
	| CAccelerator
	| CLabel
	| CPostCommand
	| CMenu
	| CScrollRegion
	| CScrollIncrement
	| CConfine
	| CCloseEnough
	| CColormap
	| CColormode
	| CFontmap
	| CPageAnchor
	| CPageHeight
	| CPageWidth
	| CPageX
	| CPageY
	| CRotate
	| CX
	| CY
	| CWindow
	| CJustify
	| CArrowStyle
	| CArrowShape
	| CCapStyle
	| CJoinStyle
	| CSmooth
	| CSplineSteps
	| CExtent
	| CFillColor
	| COutline
	| CStart
	| CStipple
	| CArcStyle
	| CTags
	| CTextWidth
	| CValue
	| CVariable
	| CSelector
	| COnValue
	| COffValue
	| CState
	| CCommand
	| CClass
	| CWidth
	| CHeight
	| CActiveBackground
	| CActiveBorderWidth
	| CActiveForeground
	| CAnchor
	| CBackground
	| CBitmap
	| CBorderwidth
	| CCursor
	| CCursorBackground
	| CCursorBorderWidth
	| CCursorOffTime
	| CCursorOnTime
	| CCursorWidth
	| CDisabledForeground
	| CExportSelection
	| CFont
	| CForeground
	| CGeometry
	| CInsertBackground
	| CInsertBorderWidth
	| CInsertOffTime
	| CInsertOnTime
	| CInsertWidth
	| COrientation
	| CPadX
	| CPadY
	| CRelief
	| CRepeatDelay
	| CRepeatInterval
	| CScrollCommand
	| CSelectBackground
	| CSelectBorderWidth
	| CSelectForeground
	| CSetGrid
	| CText
	| CTextVariable
	| CUnderlinedChar
	| CXScrollCommand
	| CYScrollCommand
;;

let option_any_table = [CRelX; CRelY; CRelWidth; CRelHeigth; CBorderMode; CAfter; CBefore; CExpand; CFill; CIn; CIPadX; CIPadY; CSide; CScreen; CWrap; CTextHeight; CBgStipple; CFgStipple; CUnderline; CTo; CTickInterval; CSliderLength; CSliderForeground; CShowValue; CLength; CFrom; CSlidecommand; CAspect; CAccelerator; CLabel; CPostCommand; CMenu; CScrollRegion; CScrollIncrement; CConfine; CCloseEnough; CColormap; CColormode; CFontmap; CPageAnchor; CPageHeight; CPageWidth; CPageX; CPageY; CRotate; CX; CY; CWindow; CJustify; CArrowStyle; CArrowShape; CCapStyle; CJoinStyle; CSmooth; CSplineSteps; CExtent; CFillColor; COutline; CStart; CStipple; CArcStyle; CTags; CTextWidth; CValue; CVariable; CSelector; COnValue; COffValue; CState; CCommand; CClass; CWidth; CHeight; CActiveBackground; CActiveBorderWidth; CActiveForeground; CAnchor; CBackground; CBitmap; CBorderwidth; CCursor; CCursorBackground; CCursorBorderWidth; CCursorOffTime; CCursorOnTime; CCursorWidth; CDisabledForeground; CExportSelection; CFont; CForeground; CGeometry; CInsertBackground; CInsertBorderWidth; CInsertOffTime; CInsertOnTime; CInsertWidth; COrientation; CPadX; CPadY; CRelief; CRepeatDelay; CRepeatInterval; CScrollCommand; CSelectBackground; CSelectBorderWidth; CSelectForeground; CSetGrid; CText; CTextVariable; CUnderlinedChar; CXScrollCommand; CYScrollCommand]
;;
let option_place_table = [CIn; CX; CY; CAnchor; CWidth; CHeight; CRelX; CRelY; CRelWidth; CRelHeigth; CBorderMode]
;;
let option_pack_table = [CAfter; CAnchor; CBefore; CExpand; CFill; CIn; CPadX; CPadY; CIPadX; CIPadY; CSide]
;;
let option_toplevel_table = [CClass; CScreen; CRelief; CGeometry; CBorderwidth; CBackground]
;;
let option_text_table = [CWrap; CTextWidth; CState; CTextHeight; CYScrollCommand; CSetGrid; CSelectForeground; CSelectBorderWidth; CSelectBackground; CRelief; CPadY; CPadX; CInsertWidth; CInsertOnTime; CInsertOffTime; CInsertBorderWidth; CInsertBackground; CForeground; CFont; CExportSelection; CCursor; CBorderwidth; CBackground]
;;
let option_texttag_table = [CBackground; CBgStipple; CBorderwidth; CFgStipple; CFont; CForeground; CRelief; CUnderline]
;;
let option_scrollbar_table = [CWidth; CSlidecommand; CRepeatInterval; CRepeatDelay; CRelief; COrientation; CForeground; CCursor; CBorderwidth; CBackground; CActiveForeground]
;;
let option_scale_table = [CWidth; CTo; CTickInterval; CState; CSliderLength; CSliderForeground; CShowValue; CLength; CLabel; CFrom; CSlidecommand; CRelief; COrientation; CForeground; CFont; CCursor; CBorderwidth; CBackground; CActiveForeground]
;;
let option_message_table = [CJustify; CAspect; CWidth; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CCursor; CBorderwidth; CBackground; CAnchor]
;;
let option_menucommand_table = [CActiveBackground; CAccelerator; CBackground; CBitmap; CCommand; CFont; CLabel; CState; CUnderlinedChar]
;;
let option_menucheck_table = [CActiveBackground; CAccelerator; CBackground; CBitmap; CCommand; CFont; CLabel; CState; CUnderlinedChar; COffValue; COnValue; CVariable]
;;
let option_menuradio_table = [CActiveBackground; CAccelerator; CBackground; CBitmap; CCommand; CFont; CLabel; CState; CUnderlinedChar; CValue; CVariable]
;;
let option_menucascade_table = [CActiveBackground; CAccelerator; CBackground; CBitmap; CCommand; CFont; CLabel; CMenu; CState; CUnderlinedChar]
;;
let option_menu_table = [CSelector; CPostCommand; CForeground; CFont; CDisabledForeground; CCursor; CBorderwidth; CBackground; CActiveForeground; CActiveBorderWidth; CActiveBackground]
;;
let option_menubutton_table = [CWidth; CState; CMenu; CHeight; CUnderlinedChar; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CDisabledForeground; CCursor; CBorderwidth; CBitmap; CBackground; CAnchor; CActiveForeground; CActiveBackground]
;;
let option_listbox_table = [CYScrollCommand; CXScrollCommand; CSetGrid; CSelectForeground; CSelectBorderWidth; CSelectBackground; CRelief; CGeometry; CFont; CForeground; CExportSelection; CCursor; CBorderwidth; CBackground]
;;
let option_label_table = [CTextWidth; CWidth; CHeight; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CCursor; CBorderwidth; CBitmap; CBackground; CAnchor]
;;
let option_canvas_table = [CWidth; CScrollRegion; CScrollIncrement; CHeight; CConfine; CCloseEnough; CYScrollCommand; CXScrollCommand; CSelectForeground; CSelectBorderWidth; CSelectBackground; CRelief; CInsertWidth; CInsertOnTime; CInsertOffTime; CInsertBorderWidth; CInsertBackground; CCursor; CBorderwidth; CBackground]
;;
let option_postscript_table = [CColormap; CColormode; CFontmap; CHeight; CPageAnchor; CPageHeight; CPageWidth; CPageX; CPageY; CRotate; CWidth; CX; CY]
;;
let option_window_table = [CAnchor; CHeight; CTags; CWidth; CWindow]
;;
let option_canvastext_table = [CAnchor; CFillColor; CFont; CJustify; CStipple; CTags; CText; CWidth]
;;
let option_rectangle_table = [CFillColor; COutline; CStipple; CTags; CWidth]
;;
let option_polygon_table = [CFillColor; CSmooth; CSplineSteps; CStipple; CTags]
;;
let option_oval_table = [CFillColor; COutline; CStipple; CTags; CWidth]
;;
let option_line_table = [CArrowStyle; CArrowShape; CCapStyle; CFillColor; CJoinStyle; CSmooth; CSplineSteps; CStipple; CTags; CWidth]
;;
let option_bitmap_table = [CAnchor; CBackground; CBitmap; CForeground; CTags]
;;
let option_arc_table = [CExtent; CFillColor; COutline; CStart; CStipple; CArcStyle; CTags; CWidth]
;;
let option_entry_table = [CTextWidth; CState; CTextVariable; CSelectForeground; CSelectBorderWidth; CSelectBackground; CScrollCommand; CRelief; CInsertWidth; CInsertOnTime; CInsertOffTime; CInsertBorderWidth; CInsertBackground; CForeground; CFont; CExportSelection; CCursor; CBorderwidth; CBackground]
;;
let option_radiobutton_table = [CWidth; CVariable; CValue; CState; CSelector; CHeight; CCommand; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CDisabledForeground; CCursor; CBorderwidth; CBitmap; CBackground; CAnchor; CActiveForeground; CActiveBackground]
;;
let option_checkbutton_table = [CVariable; CWidth; CState; CSelector; COnValue; COffValue; CHeight; CCommand; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CDisabledForeground; CCursor; CBorderwidth; CBitmap; CBackground; CAnchor; CActiveForeground; CActiveBackground]
;;
let option_button_table = [CWidth; CState; CHeight; CCommand; CTextVariable; CText; CRelief; CPadY; CPadX; CForeground; CFont; CDisabledForeground; CCursor; CBorderwidth; CBitmap; CBackground; CAnchor; CActiveForeground; CActiveBackground]
;;
let option_frame_table = [CClass; CWidth; CHeight; CRelief; CGeometry; CCursor; CBorderwidth; CBackground]
;;
let option_standard_table = [CActiveBackground; CActiveBorderWidth; CActiveForeground; CAnchor; CBackground; CBitmap; CBorderwidth; CCursor; CCursorBackground; CCursorBorderWidth; CCursorOffTime; CCursorOnTime; CCursorWidth; CDisabledForeground; CExportSelection; CFont; CForeground; CGeometry; CInsertBackground; CInsertBorderWidth; CInsertOffTime; CInsertOnTime; CInsertWidth; COrientation; CPadX; CPadY; CRelief; CRepeatDelay; CRepeatInterval; CScrollCommand; CSelectBackground; CSelectBorderWidth; CSelectForeground; CSetGrid; CText; CTextVariable; CUnderlinedChar; CXScrollCommand; CYScrollCommand]
;;
let CAMLtoTKoption table = function
	RelX x -> chk_sub "RelX" table CRelX; "-relx"^" "^CAMLtoTKUnits x
	| RelY x -> chk_sub "RelY" table CRelY; "-rely"^" "^CAMLtoTKUnits x
	| RelWidth x -> chk_sub "RelWidth" table CRelWidth; "-relwidth"^" "^string_of_float x
	| RelHeigth x -> chk_sub "RelHeigth" table CRelHeigth; "-relheight"^" "^string_of_float x
	| BorderMode x -> chk_sub "BorderMode" table CBorderMode; "-bordermode"^" "^CAMLtoTKBorderMode x
	| After x -> chk_sub "After" table CAfter; "-after"^" "^CAMLtoTKWidget Widget_any_table x
	| Before x -> chk_sub "Before" table CBefore; "-before"^" "^CAMLtoTKWidget Widget_any_table x
	| Expand x -> chk_sub "Expand" table CExpand; "-expand"^" "^if x then "1" else "0"
	| Fill x -> chk_sub "Fill" table CFill; "-fill"^" "^CAMLtoTKFillMode x
	| In x -> chk_sub "In" table CIn; "-in"^" "^CAMLtoTKWidget Widget_any_table x
	| IPadX x -> chk_sub "IPadX" table CIPadX; "-ipadx"^" "^CAMLtoTKUnits x
	| IPadY x -> chk_sub "IPadY" table CIPadY; "-ipady"^" "^CAMLtoTKUnits x
	| Side x -> chk_sub "Side" table CSide; "-side"^" "^CAMLtoTKSide x
	| Screen x -> chk_sub "Screen" table CScreen; "-screen"^" "^quote_string x
	| Wrap x -> chk_sub "Wrap" table CWrap; "-wrap"^" "^CAMLtoTKWrapMode x
	| TextHeight x -> chk_sub "TextHeight" table CTextHeight; "-height"^" "^string_of_int x
	| BgStipple x -> chk_sub "BgStipple" table CBgStipple; "-bgstipple"^" "^CAMLtoTKBitmap x
	| FgStipple x -> chk_sub "FgStipple" table CFgStipple; "-fgstipple"^" "^CAMLtoTKBitmap x
	| Underline x -> chk_sub "Underline" table CUnderline; "-underline"^" "^if x then "1" else "0"
	| To x -> chk_sub "To" table CTo; "-to"^" "^string_of_int x
	| TickInterval x -> chk_sub "TickInterval" table CTickInterval; "-tickinterval"^" "^string_of_int x
	| SliderLength x -> chk_sub "SliderLength" table CSliderLength; "-sliderlength"^" "^CAMLtoTKUnits x
	| SliderForeground x -> chk_sub "SliderForeground" table CSliderForeground; "-sliderforeground"^" "^CAMLtoTKColor x
	| ShowValue x -> chk_sub "ShowValue" table CShowValue; "-showvalue"^" "^if x then "1" else "0"
	| Length x -> chk_sub "Length" table CLength; "-length"^" "^CAMLtoTKUnits x
	| From x -> chk_sub "From" table CFrom; "-from"^" "^string_of_int x
	| Slidecommand f -> chk_sub "Slidecommand" table CSlidecommand; "-command"^" "^let id = register_callback (function args ->
		f(int_of_string (arg_GetTkToken args))) in "{camlcb "^id^"}"
	| Aspect x -> chk_sub "Aspect" table CAspect; "-aspect"^" "^string_of_int x
	| Accelerator x -> chk_sub "Accelerator" table CAccelerator; "-accelerator"^" "^quote_string x
	| Label x -> chk_sub "Label" table CLabel; "-label"^" "^quote_string x
	| PostCommand f -> chk_sub "PostCommand" table CPostCommand; "-postcommand"^" "^let id = register_callback (function _ -> f ()) in "{camlcb "^id^"}"
	| Menu x -> chk_sub "Menu" table CMenu; "-menu"^" "^CAMLtoTKWidget Widget_menu_table x
	| ScrollRegion( a1, a2, a3, a4) -> chk_sub "ScrollRegion" table CScrollRegion; "-scrollregion"^" "^CAMLtoTKUnits a1^" "^CAMLtoTKUnits a2^" "^CAMLtoTKUnits a3^" "^CAMLtoTKUnits a4
	| ScrollIncrement x -> chk_sub "ScrollIncrement" table CScrollIncrement; "-scrollincrement"^" "^CAMLtoTKUnits x
	| Confine x -> chk_sub "Confine" table CConfine; "-confine"^" "^if x then "1" else "0"
	| CloseEnough x -> chk_sub "CloseEnough" table CCloseEnough; "-closeenough"^" "^string_of_float x
	| Colormap x -> chk_sub "Colormap" table CColormap; "-colormap"^" "^quote_string x
	| Colormode x -> chk_sub "Colormode" table CColormode; "-colormode"^" "^CAMLtoTKColorMode x
	| Fontmap x -> chk_sub "Fontmap" table CFontmap; "-fontmap"^" "^quote_string x
	| PageAnchor x -> chk_sub "PageAnchor" table CPageAnchor; "-pageanchor"^" "^CAMLtoTKAnchor x
	| PageHeight x -> chk_sub "PageHeight" table CPageHeight; "-pageheight"^" "^CAMLtoTKUnits x
	| PageWidth x -> chk_sub "PageWidth" table CPageWidth; "-pagewidth"^" "^CAMLtoTKUnits x
	| PageX x -> chk_sub "PageX" table CPageX; "-pagex"^" "^CAMLtoTKUnits x
	| PageY x -> chk_sub "PageY" table CPageY; "-pagey"^" "^CAMLtoTKUnits x
	| Rotate x -> chk_sub "Rotate" table CRotate; "-rotate"^" "^if x then "1" else "0"
	| X x -> chk_sub "X" table CX; "-x"^" "^CAMLtoTKUnits x
	| Y x -> chk_sub "Y" table CY; "-y"^" "^CAMLtoTKUnits x
	| Window x -> chk_sub "Window" table CWindow; "-window"^" "^CAMLtoTKWidget Widget_any_table x
	| Justify x -> chk_sub "Justify" table CJustify; "-justify"^" "^CAMLtoTKJustification x
	| ArrowStyle x -> chk_sub "ArrowStyle" table CArrowStyle; "-arrow"^" "^CAMLtoTKArrowStyle x
	| ArrowShape( a1, a2, a3) -> chk_sub "ArrowShape" table CArrowShape; "-arrowshape"^" "^CAMLtoTKUnits a1^" "^CAMLtoTKUnits a2^" "^CAMLtoTKUnits a3
	| CapStyle x -> chk_sub "CapStyle" table CCapStyle; "-capstyle"^" "^CAMLtoTKCapStyle x
	| JoinStyle x -> chk_sub "JoinStyle" table CJoinStyle; "-joinstyle"^" "^CAMLtoTKJoinStyle x
	| Smooth x -> chk_sub "Smooth" table CSmooth; "-smooth"^" "^if x then "1" else "0"
	| SplineSteps x -> chk_sub "SplineSteps" table CSplineSteps; "-splinesteps"^" "^string_of_int x
	| Extent x -> chk_sub "Extent" table CExtent; "-extent"^" "^string_of_float x
	| FillColor x -> chk_sub "FillColor" table CFillColor; "-fill"^" "^CAMLtoTKColor x
	| Outline x -> chk_sub "Outline" table COutline; "-outline"^" "^CAMLtoTKColor x
	| Start x -> chk_sub "Start" table CStart; "-start"^" "^string_of_float x
	| Stipple x -> chk_sub "Stipple" table CStipple; "-stipple"^" "^CAMLtoTKBitmap x
	| ArcStyle x -> chk_sub "ArcStyle" table CArcStyle; "-style"^" "^CAMLtoTKArcStyle x
	| Tags x -> chk_sub "Tags" table CTags; "-tags"^" "^catenate_sep " " (map (function x -> CAMLtoTKTagOrId x) x)
	| TextWidth x -> chk_sub "TextWidth" table CTextWidth; "-width"^" "^string_of_int x
	| Value x -> chk_sub "Value" table CValue; "-value"^" "^quote_string x
	| Variable x -> chk_sub "Variable" table CVariable; "-variable"^" "^CAMLtoTKTextVariable x
	| Selector x -> chk_sub "Selector" table CSelector; "-selector"^" "^CAMLtoTKColor x
	| OnValue x -> chk_sub "OnValue" table COnValue; "-onvalue"^" "^quote_string x
	| OffValue x -> chk_sub "OffValue" table COffValue; "-offvalue"^" "^quote_string x
	| State x -> chk_sub "State" table CState; "-state"^" "^CAMLtoTKState x
	| Command f -> chk_sub "Command" table CCommand; "-command"^" "^let id = register_callback (function _ -> f ()) in "{camlcb "^id^"}"
	| Class x -> chk_sub "Class" table CClass; "-class"^" "^quote_string x
	| Width x -> chk_sub "Width" table CWidth; "-width"^" "^CAMLtoTKUnits x
	| Height x -> chk_sub "Height" table CHeight; "-height"^" "^CAMLtoTKUnits x
	| ActiveBackground x -> chk_sub "ActiveBackground" table CActiveBackground; "-activebackground"^" "^CAMLtoTKColor x
	| ActiveBorderWidth x -> chk_sub "ActiveBorderWidth" table CActiveBorderWidth; "-activeborderwidth"^" "^CAMLtoTKUnits x
	| ActiveForeground x -> chk_sub "ActiveForeground" table CActiveForeground; "-activeforeground"^" "^CAMLtoTKColor x
	| Anchor x -> chk_sub "Anchor" table CAnchor; "-anchor"^" "^CAMLtoTKAnchor x
	| Background x -> chk_sub "Background" table CBackground; "-background"^" "^CAMLtoTKColor x
	| Bitmap x -> chk_sub "Bitmap" table CBitmap; "-bitmap"^" "^CAMLtoTKBitmap x
	| Borderwidth x -> chk_sub "Borderwidth" table CBorderwidth; "-borderwidth"^" "^CAMLtoTKUnits x
	| Cursor x -> chk_sub "Cursor" table CCursor; "-cursor"^" "^CAMLtoTKCursor x
	| CursorBackground x -> chk_sub "CursorBackground" table CCursorBackground; "-cursorbackground"^" "^CAMLtoTKColor x
	| CursorBorderWidth x -> chk_sub "CursorBorderWidth" table CCursorBorderWidth; "-cursorborderwidth"^" "^CAMLtoTKUnits x
	| CursorOffTime x -> chk_sub "CursorOffTime" table CCursorOffTime; "-cursorofftime"^" "^string_of_int x
	| CursorOnTime x -> chk_sub "CursorOnTime" table CCursorOnTime; "-cursorontime"^" "^string_of_int x
	| CursorWidth x -> chk_sub "CursorWidth" table CCursorWidth; "-cursorwidth"^" "^CAMLtoTKUnits x
	| DisabledForeground x -> chk_sub "DisabledForeground" table CDisabledForeground; "-disabledforeground"^" "^CAMLtoTKColor x
	| ExportSelection x -> chk_sub "ExportSelection" table CExportSelection; "-exportselection"^" "^if x then "1" else "0"
	| Font x -> chk_sub "Font" table CFont; "-font"^" "^quote_string x
	| Foreground x -> chk_sub "Foreground" table CForeground; "-foreground"^" "^CAMLtoTKColor x
	| Geometry x -> chk_sub "Geometry" table CGeometry; "-geometry"^" "^quote_string x
	| InsertBackground x -> chk_sub "InsertBackground" table CInsertBackground; "-insertbackground"^" "^CAMLtoTKColor x
	| InsertBorderWidth x -> chk_sub "InsertBorderWidth" table CInsertBorderWidth; "-insertborderwidth"^" "^CAMLtoTKUnits x
	| InsertOffTime x -> chk_sub "InsertOffTime" table CInsertOffTime; "-insertofftime"^" "^string_of_int x
	| InsertOnTime x -> chk_sub "InsertOnTime" table CInsertOnTime; "-insertontime"^" "^string_of_int x
	| InsertWidth x -> chk_sub "InsertWidth" table CInsertWidth; "-insertwidth"^" "^CAMLtoTKUnits x
	| Orientation x -> chk_sub "Orientation" table COrientation; "-orient"^" "^CAMLtoTKOrientation x
	| PadX x -> chk_sub "PadX" table CPadX; "-padx"^" "^CAMLtoTKUnits x
	| PadY x -> chk_sub "PadY" table CPadY; "-pady"^" "^CAMLtoTKUnits x
	| Relief x -> chk_sub "Relief" table CRelief; "-relief"^" "^CAMLtoTKRelief x
	| RepeatDelay x -> chk_sub "RepeatDelay" table CRepeatDelay; "-repeatdelay"^" "^string_of_int x
	| RepeatInterval x -> chk_sub "RepeatInterval" table CRepeatInterval; "-repeatinterval"^" "^string_of_int x
	| ScrollCommand f -> chk_sub "ScrollCommand" table CScrollCommand; "-scrollcommand"^" "^let id = register_callback (function args ->
		let a1 = int_of_string (arg_GetTkToken args) in
		let a2 = int_of_string (arg_GetTkToken args) in
		let a3 = int_of_string (arg_GetTkToken args) in
		let a4 = int_of_string (arg_GetTkToken args) in
		f a1 a2 a3 a4) in "{camlcb "^id^"}"
	| SelectBackground x -> chk_sub "SelectBackground" table CSelectBackground; "-selectbackground"^" "^CAMLtoTKColor x
	| SelectBorderWidth x -> chk_sub "SelectBorderWidth" table CSelectBorderWidth; "-selectborderwidth"^" "^CAMLtoTKUnits x
	| SelectForeground x -> chk_sub "SelectForeground" table CSelectForeground; "-selectforeground"^" "^CAMLtoTKColor x
	| SetGrid x -> chk_sub "SetGrid" table CSetGrid; "-setgrid"^" "^if x then "1" else "0"
	| Text x -> chk_sub "Text" table CText; "-text"^" "^quote_string x
	| TextVariable x -> chk_sub "TextVariable" table CTextVariable; "-textvariable"^" "^CAMLtoTKTextVariable x
	| UnderlinedChar x -> chk_sub "UnderlinedChar" table CUnderlinedChar; "-underline"^" "^string_of_int x
	| XScrollCommand f -> chk_sub "XScrollCommand" table CXScrollCommand; "-xscrollcommand"^" "^let id = register_callback (function args ->
		let a1 = int_of_string (arg_GetTkToken args) in
		let a2 = int_of_string (arg_GetTkToken args) in
		let a3 = int_of_string (arg_GetTkToken args) in
		let a4 = int_of_string (arg_GetTkToken args) in
		f a1 a2 a3 a4) in "{camlcb "^id^"}"
	| YScrollCommand f -> chk_sub "YScrollCommand" table CYScrollCommand; "-yscrollcommand"^" "^let id = register_callback (function args ->
		let a1 = int_of_string (arg_GetTkToken args) in
		let a2 = int_of_string (arg_GetTkToken args) in
		let a3 = int_of_string (arg_GetTkToken args) in
		let a4 = int_of_string (arg_GetTkToken args) in
		f a1 a2 a3 a4) in "{camlcb "^id^"}"
;;

type Index =
	Number of int
	| End
	| Insert
	| SelFirst
	| SelLast
;;

type Index_constrs =
	CNumber
	| CEnd
	| CInsert
	| CSelFirst
	| CSelLast
;;

let Index_any_table = [CNumber; CEnd; CInsert; CSelFirst; CSelLast]
;;
let Index_listboxget_table = [CNumber; CEnd]
;;
let Index_listbox_table = [CNumber; CEnd; CInsert]
;;
let Index_entry_table = [CNumber; CEnd; CInsert; CSelFirst; CSelLast]
;;
let CAMLtoTKIndex table = function
	Number x -> chk_sub "Number" table CNumber; string_of_int x
	| End -> chk_sub "End" table CEnd; "end"
	| Insert -> chk_sub "Insert" table CInsert; "insert"
	| SelFirst -> chk_sub "SelFirst" table CSelFirst; "sel.first"
	| SelLast -> chk_sub "SelLast" table CSelLast; "sel.last"
;;

let TKtoCAMLIndex n =
   try Number(int_of_string n)
   with _ ->
	match n with
	"sel.last" -> SelLast
	| "sel.first" -> SelFirst
	| "insert" -> Insert
	| "end" -> End
	|  _ -> raise (Invalid_argument "TKtoCAMLIndex")
;;
type CanvasItem =
	Window_item
	| Text_item
	| Polygon_item
	| Bitmap_item
	| Oval_item
	| Rectangle_item
	| Line_item
	| Arc_item
;;

let CAMLtoTKCanvasItem = function
	Window_item -> "window"
	| Text_item -> "text"
	| Polygon_item -> "polygon"
	| Bitmap_item -> "bitmap"
	| Oval_item -> "oval"
	| Rectangle_item -> "rectangle"
	| Line_item -> "line"
	| Arc_item -> "arc"
;;

let TKtoCAMLCanvasItem n =
	match n with
	"arc" -> Arc_item
	| "line" -> Line_item
	| "rectangle" -> Rectangle_item
	| "oval" -> Oval_item
	| "bitmap" -> Bitmap_item
	| "polygon" -> Polygon_item
	| "text" -> Text_item
	| "window" -> Window_item
	|  _ -> raise (Invalid_argument "TKtoCAMLCanvasItem")
;;
type Comparison =
	NEQ
	| LT
	| LE
	| EQ
	| GE
	| GT
;;

let CAMLtoTKComparison = function
	NEQ -> "!="
	| LT -> ">"
	| LE -> "=>"
	| EQ -> "=="
	| GE -> "<="
	| GT -> "<"
;;

type GrabStatus =
	GrabGlobal
	| GrabLocal
	| GrabNone
;;

let CAMLtoTKGrabStatus = function
	GrabGlobal -> "global"
	| GrabLocal -> "local"
	| GrabNone -> "none"
;;

let TKtoCAMLGrabStatus n =
	match n with
	"none" -> GrabNone
	| "local" -> GrabLocal
	| "global" -> GrabGlobal
	|  _ -> raise (Invalid_argument "TKtoCAMLGrabStatus")
;;
type OptionPriority =
	Interactive
	| UserDefault
	| StartupFile
	| WidgetDefault
;;

let CAMLtoTKOptionPriority = function
	Interactive -> "interactive"
	| UserDefault -> "userDefault"
	| StartupFile -> "startupFile"
	| WidgetDefault -> "widgetDefault"
;;

type ColorModel =
	Monochrome
	| Color
;;

let CAMLtoTKColorModel = function
	Monochrome -> "monochrome"
	| Color -> "color"
;;

let TKtoCAMLColorModel n =
	match n with
	"color" -> Color
	| "monochrome" -> Monochrome
	|  _ -> raise (Invalid_argument "TKtoCAMLColorModel")
;;
type AtomId =
	AtomId of int
;;

let CAMLtoTKAtomId = function
	AtomId x -> string_of_int x
;;

let TKtoCAMLAtomId n =
   try AtomId(int_of_string n)
   with _ ->
	match n with
	 _ -> raise (Invalid_argument "TKtoCAMLAtomId")
;;
type FocusModel =
	FocusPassive
	| FocusActive
;;

let CAMLtoTKFocusModel = function
	FocusPassive -> "passive"
	| FocusActive -> "active"
;;

let TKtoCAMLFocusModel n =
	match n with
	"active" -> FocusActive
	| "passive" -> FocusPassive
	|  _ -> raise (Invalid_argument "TKtoCAMLFocusModel")
;;
let update_idletasks () =
	let buf = Send2TkStart false in 
	Send2Tk buf "update idletasks";
	Send2TkEval buf; ()
;;

let update () =
	let buf = Send2TkStart false in 
	Send2Tk buf "update";
	Send2TkEval buf; ()
;;

let place a1 a2 =
	let buf = Send2TkStart false in 
	Send2Tk buf "place";
	Send2Tk buf (CAMLtoTKWidget Widget_any_table a1);
	Send2Tk buf (catenate_sep " " (map (function x -> CAMLtoTKoption option_place_table x) a2));
	Send2TkEval buf; ()
;;

let pack a1 a2 =
	let buf = Send2TkStart false in 
	Send2Tk buf "pack";
	Send2Tk buf (catenate_sep " " (map (function x -> CAMLtoTKWidget Widget_any_table x) a1));
	Send2Tk buf (catenate_sep " " (map (function x -> CAMLtoTKoption option_pack_table x) a2));
	Send2TkEval buf; ()
;;

let lower a =
	let buf = Send2TkStart false in 
	Send2Tk buf "lower";
	Send2Tk buf (CAMLtoTKWidget Widget_any_table a);
	Send2TkEval buf; ()
;;

let dialog a1 a2 a3 a4 a5 a6 =
	let buf = Send2TkStart true in 
	Send2Tk buf result_header;
	Send2Tk buf "tk_dialog";
	Send2Tk buf (CAMLtoTKWidget Widget_any_table a1);
	Send2Tk buf (quote_string a2);
	Send2Tk buf (quote_string a3);
	Send2Tk buf (quote_string a4);
	Send2Tk buf (string_of_int a5);
	Send2Tk buf (catenate_sep " " (map (function x -> quote_string x) a6));
	Send2Tk buf result_footer;
	let res = Send2TkEval buf in
	int_of_string (res_GetTkToken res);;

let destroy a =
	let buf = Send2TkStart false in 
	Send2Tk buf "destroy";
	Send2Tk buf (CAMLtoTKWidget Widget_any_table a);
	Send2TkEval buf; ()
;;

